"""
Simple implementation of the CulliganIoT API

This is used to provide an interface between Culligan and Ayla. 
Additional API endpoints are unknown and closed source.

Logging in directly to Ayla still works for some endpoints. 
Others such as /devices.json only seem to work when obtaining the token through Culligan.
"""

from aiohttp    import ClientSession             # async http
from ayla_iot_unofficial import AylaApi
from .culliganiot_device import CulliganIoTDevice, CulliganIoTSoftener
from datetime   import datetime, timedelta       # datetime operations
from requests   import post, put, request, Response   # http request library
from typing     import Dict, List, Optional      # object types

try:
    from ujson   import loads
except ImportError:
    from json    import loads

# Defined constants
from .const import (
    CULLIGAN_APP_ID,
    CULLIGAN_IOT_URL
)

# Custom error handling 
from .exc import (
    CulliganError,
    CulliganAuthExpiringError,
    CulliganNotAuthedError,
    CulliganAuthError,
    CulliganReadOnlyPropertyError,
)

_session = None

def new_culligan_api(username: str, password: str, app_id: str = CULLIGAN_APP_ID, websession: Optional[ClientSession] = None):
    """Get an CulliganApi object. Username is an email address."""
    return CulliganApi(username, password, app_id, websession=websession)


class CulliganApi:
    """Simple CulliganIoT Networks API wrapper"""

    def __init__(
            self,
            username: str,
            password: str,
            app_id: str,
            websession: Optional[ClientSession] = None):
        self._email                 = username      # username should always be an email address
        self._password              = password
        self._culligan_username     = None
        self._culligan_access_token = None
        self._culligan_refresh_token= None
        self._culligan_expiration   = None
        self._ayla_access_token     = None          # type: Optional[str]
        self._ayla_refresh_token    = None          # type: Optional[str]
        self._ayla_expiration       = None          # type: Optional[datetime]
        self._ayla_expiration_raw   = None          # type: Optional[int]
        self._is_authed             = False         # type: bool
        self._app_id                = app_id
        self.Ayla                   = None
        self.websession             = websession
        self.v1_url                 = CULLIGAN_IOT_URL

        # identify serials and dsns to track
        self.culligan_iot_serials   = []
        self.ayla_networks_dsns     = None

        # eventually combine them using Ayla 'softener' devices or other generic device class
        # ayla Softener(Device) class updates differently ... maybe extend Device here in Culligan

    async def ensure_session(self) -> ClientSession:
        """Ensure that we have an aiohttp ClientSession"""
        if self.websession is None:
            self.websession = ClientSession()
        return self.websession

    @property
    def _login_data(self) -> Dict[str, Dict]:
        """Prettily formatted data for the login flow"""
        return {
            "email": self._email,
            "password": self._password,
            "appId": self._app_id
        }

    @property
    def _sign_out_data(self) -> Dict:
        """Payload for the sign_out call"""
        return {"access_token": self._culligan_access_token}
    
    @property
    def _refresh_data(self) -> Dict:
        """Payload for refresh token call. Refresh tokens are currently long-lived and do not seem to change."""
        return {
            "refreshToken": self._culligan_refresh_token, 
            "appId": CULLIGAN_APP_ID
        }

    def _set_credentials(self, status_code: int, login_result: Dict):
        """Update the internal credentials store. This tracks current bearer token and data needed for token refresh."""
        if status_code   == 404:
            raise CulliganAuthError(login_result["error"]["message"] + " (Confirm login information is correct)")
        elif status_code == 401:
            raise CulliganAuthError(login_result["error"]["message"])
        elif status_code == 422:
            raise CulliganAuthError(login_result["error"]["message"] + " (Confirm login information is correct, username should be an email address.)")
        elif "data" not in login_result:
            raise CulliganAuthError(login_result + "Something unexpected happened and there was no 'data' in the response.")

        self._culligan_username     = login_result["data"]["userId"]
        self._culligan_access_token = login_result["data"]["accessToken"]
        self._culligan_refresh_token= login_result["data"]["refreshToken"]
        self._culligan_expiration   = datetime.now() + timedelta(seconds=login_result["data"]["expiresIn"])

        # Ayla tokens are not guaranteed to exist ... 
        if "ayla" in login_result["data"]["linkedAccounts"]:
            self._ayla_access_token     = login_result["data"]["linkedAccounts"]["ayla"]["access_token"]
            self._ayla_refresh_token    = login_result["data"]["linkedAccounts"]["ayla"]["refresh_token"]
            self._ayla_expiration       = datetime.now() + timedelta(seconds=login_result["data"]["linkedAccounts"]["ayla"]["expires_in"])
            self._ayla_expiration_raw   = login_result["data"]["linkedAccounts"]["ayla"]["expires_in"]

        if status_code != 200:
            self._is_authed   = False
        else:
            self._is_authed   = True

    def sign_in(self, returnResponse: bool = False):
        """ 
            Authenticate to Culligan API synchronously using a POST with credentials.
            Culligan may continue to hide/remove interactions with Ayla directly in the future.
                e.g. app_secret is no longer available and the previous app_secret no longer populates devices.json
            In the meantime, Culligan provides the ayla access_token directly to the user for direct use.
            To obtain an Ayla specific object (e.g. ayla_iot_unofficial AylaApi class object): 
            use returnResponse = True
            instantiate AylaApi
            then call AylaApi._set_credentials(200,resp["data"]["linkedAccounts"]["ayla"])
        """
        login_data = self._login_data   # get a map for JSON formatting
        resp = post(f"{self.v1_url:s}/auth/login", json=login_data)
        self._set_credentials(resp.status_code, resp.json())
        if returnResponse:
            return resp.json()

    def refresh_auth(self):
        """Refresh the authentication synchronously using object tracked refresh token."""
        refresh_data = self._refresh_data
        resp = put(f"{self.v1_url:s}/auth/login", json=refresh_data)
        self._set_credentials(resp.status_code, resp.json())

    async def async_sign_in(self):
        """Authenticate to Culligan API asynchronously using a POST with credentials.."""
        session = await self.ensure_session()
        login_data = self._login_data
        async with session.post(f"{self.v1_url:s}/auth/login", json=login_data) as resp:
            self._set_credentials(resp.status, await resp.json())

    async def async_refresh_auth(self):
        """Refresh the authentication asynchronously using object tracked refresh token.."""
        session = await self.ensure_session()
        refresh_data = self._refresh_data
        async with session.put(f"{self.v1_url:s}/auth/login", json=refresh_data) as resp:
            self._set_credentials(resp.status, await resp.json())

    def _clear_auth(self):
        """Clear authentication state"""
        self._is_authed             = False
        self._culligan_access_token = None
        self._culligan_refresh_token= None
        self._culligan_expiration   = None
        self._ayla_access_token     = None
        self._ayla_refresh_token    = None
        self._ayla_expiration       = None
        self.websession             = None

    def sign_out(self):
        """Sign out and invalidate the access token synchronously"""
        """Placeholder until logout is known"""
        #post(f"{self.v1_url:s}/auth/login", json=self._sign_out_data)
        self._clear_auth()

    async def async_sign_out(self):
        """Sign out and invalidate the access token asynchronously"""
        """Placeholder until logout is known"""
        session = await self.ensure_session()
        #async with session.post(f"{self.v1_url:s}/auth/login", json=self._sign_out_data) as _:
        #    pass
        #self._clear_auth()

    @property
    def auth_expiration(self) -> Optional[datetime]:
        """When does the auth expire"""
        if not self._is_authed:
            return None
        elif self._culligan_expiration is None:  # This should not happen, but let's be ready if it does...
            raise CulliganNotAuthedError("Invalid state.  Please reauthorize.")
        else:
            return self._culligan_expiration

    @property
    def token_expired(self) -> bool:
        """Return true if the token has already expired"""
        if self.auth_expiration is None:
            return True
        return datetime.now() > self.auth_expiration

    @property
    def token_expiring_soon(self) -> bool:
        """Return true if the token will expire soon"""
        if self.auth_expiration is None:
            return True
        return datetime.now() > self.auth_expiration - timedelta(seconds=600)  # Prevent timeout immediately following

    def check_auth(self, raise_expiring_soon=True):
        """Confirm authentication status"""
        if not self._culligan_access_token or not self._is_authed or self.token_expired:
            self._is_authed = False
            raise CulliganNotAuthedError()
        elif raise_expiring_soon and self.token_expiring_soon:
            raise CulliganAuthExpiringError()

    @property
    def auth_header(self) -> Dict[str, str]:
        self.check_auth()
        return {"Authorization": f"Bearer {self._culligan_access_token:s}"}
    
    @property
    def no_cache_header(self) -> Dict[str, str]:
        """If device/registry will be annoying later, can specify a force update using cache-control."""
        return {"Cache-Control": "no-cache"}

    def _get_headers(self, fn_kwargs) -> Dict[str, str]:
        """
        Extract the headers element from fn_kwargs, removing it if it exists
        and updating with self.auth_header.
        """
        try:
            headers = fn_kwargs['headers']
        except KeyError:
            headers = {}
        else:
            del fn_kwargs['headers']
        headers.update(self.auth_header)
        return headers

    def self_request(self, method: str, url: str, **kwargs) -> Response:
        """Perform an arbitrary request using the requests library synchronously"""
        headers = self._get_headers(kwargs)
        return request(method, url, headers=headers, **kwargs)

    async def async_request(self, http_method: str, url: str, **kwargs):
        """Perform an arbitrary request using the aiohttp library asynchronously"""
        session = await self.ensure_session()
        headers = self._get_headers(kwargs)
        return session.request(http_method, url, headers=headers, **kwargs)

    def get_ayla_api(self) -> AylaApi:
        """ Get an instace of the AylaApi object and force instantiate it with auth provided by Culligan """
        AuthFromCulligan = {
            "access_token": self._ayla_access_token,
            "refresh_token": self._ayla_refresh_token,
            "expires_in": self._ayla_expiration_raw
        }

        # Check for data or set to false ... Ayla not guaranteed to be a linked account
        if AuthFromCulligan["access_token"]:
            Ayla = AylaApi(self._email, self._password, self._app_id, None, self.websession, False)
            Ayla._set_credentials(200, AuthFromCulligan)
            return Ayla
        else:
            return False

    def get_user_profile(self) -> Dict[str, str]:
        """Get user profile synchronously"""
        resp = self.self_request("get", f"{self.v1_url:s}/user/profile")
        response = resp.json()
        if resp.status_code == 401:
            raise CulliganAuthError(response)
        return response
    
    async def async_get_user_profile(self) -> Dict[str, str]:
        """Get user profile async"""
        async with await self.async_request("get", f"{self.v1_url:s}/user/profile") as resp:
            response = await resp.json()
            if resp.status == 401:
                raise CulliganAuthError(response)
        return response
    
    def get_metadata_user(self) -> Dict[str, str]:
        """Get user metadata synchronously. This is the CWS onboarding survey results (house side, what's new, interests, etc)"""
        resp = self.self_request("get", f"{self.v1_url:s}/metadata/user")
        response = resp.json()
        if resp.status_code == 401:
            raise CulliganAuthError(response)
        response["data"]["CWS-onboarding-survey"] = loads(response["data"]["CWS-onboarding-survey"])
        return response
    
    async def async_get_metadata_user(self) -> Dict[str, str]:
        """Get user metadata asynchronously. This is the CWS onboarding survey results (house side, what's new, interests, etc)"""
        async with await self.async_request("get", f"{self.v1_url:s}/metadata/user") as resp:
            response = await resp.json()
            if resp.status == 401:
                raise CulliganAuthError(response)
        response["data"]["CWS-onboarding-survey"] = loads(response["data"]["CWS-onboarding-survey"])
        return response
    
    def get_device_registry(self) -> Dict[str, str]:
        """Get device registry synchronously"""
        # returns data{devices[]}
        # devices has: serialNumber, name, model, generation, protocolVersion, lat, lon, swVersion, status{connection{online, lastUpdate}}, region{code}
        # metadata{}, registeredAt, createdAt, updatedAt, currentUserrole, dealerId, accountNumber, installationAddress{address, zip, city, state, country}
        resp = self.self_request("get", f"{self.v1_url:s}/device/registry")
        response = resp.json()
        if resp.status_code == 401:
            raise CulliganAuthError(response)
        
        # track serialNumbers for device/data endpoint
        for device in response["data"]["devices"]:
            if device["serialNumber"] not in self.culligan_iot_serials:
                self.culligan_iot_serials.append(device["serialNumber"])

        return response
    
    async def async_get_device_registry(self) -> Dict[str, str]:
        """Get device registry async"""
        async with await self.async_request("get", f"{self.v1_url:s}/device/registry") as resp:
            response = await resp.json()
            if resp.status == 401:
                raise CulliganAuthError(response)
            
        # track serialNumbers for device/data endpoint
        for device in response["data"]["devices"]:
            if device["serialNumber"] not in self.culligan_iot_serials:
                self.culligan_iot_serials.append(device["serialNumber"])

        return response
    
    def get_devices(self) -> List[CulliganIoTDevice]:
        """Retrieve a device object of devices. Ability to update with metadata. Synchronous."""
        devices = list()
        device_registry = self.get_device_registry()
        for d in device_registry["data"]["devices"]:
            # Have no idea what products will be enabled or how to identify them ... for now ... Smart HE is a softener
            if   d["name"] in ["Smart HE","Smart Modernity"]:
                devices.append(CulliganIoTSoftener(self, d))
            # Everything else is a device
            else:
                devices.append(CulliganIoTDevice  (self, d))
        return devices
    
    async def async_get_devices(self) -> List[CulliganIoTDevice]:
        """Retrieve a device object of devices. Ability to update with metadata. Asynchronous."""
        devices = list()
        device_registry = await self.async_get_device_registry()
        for d in device_registry["data"]["devices"]:
            # Have no idea what products will be enabled or how to identify them ... for now ... Smart HE is a softener
            if   d["name"] in ["Smart HE","Smart Modernity"]:
                devices.append(CulliganIoTSoftener(self, d))
            # Everything else is a device
            else:
                devices.append(CulliganIoTDevice  (self, d))
        return devices

    def get_device_data(self, serialNumber: str) -> Dict[str, str]:
        """Get device registry synchronously"""
        # most ayla properties in [data][datapoints]
        resp = self.self_request("get", f"{self.v1_url:s}/device/data?serialNumber={serialNumber:s}")
        response = resp.json()
        if resp.status_code == 401:
            raise CulliganAuthError(response)
        return response
    
    async def async_get_device_data(self, serialNumber: str) -> Dict[str, str]:
        """Get device registry async"""
        # most ayla properties in [data][datapoints]
        async with await self.async_request("get", f"{self.v1_url:s}/device/data?serialNumber={serialNumber:s}") as resp:
            response = await resp.json()
            if resp.status == 401:
                raise CulliganAuthError(response)
        return response



"""Class for devices handled by CulliganIoT domain"""

from .const import CULLIGAN_IOT_URL
from enum        import Enum, IntEnum, unique
from typing      import Any, Dict, Iterable, List, Optional, Set, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from .uniapi_culliganiot import CulliganApi

#_LOGGER = getLogger(__name__)

PropertyName  = Union[str, Enum]
PropertyValue = Union[str, int, Enum]

class CulliganIoTDevice:
    """Generic device entity"""

    def __init__(self, culligan_api: "CulliganApi", device_dct: Dict):
        """
            Start object with serial = dsn. For some devices (such as SharkIQ vacuums) a device serial number is needed instead.
                call _update_metadata() to update these values.
                Other objects (such as Culligan water softeners), the DSN is used in place of serial numbers
        """
        self.culligan_api           = culligan_api

        self.properties             = {}
        self._commands              = []

        # Properties
        self._name                  = device_dct['name']
        self._device_serial_number  = device_dct["serialNumber"]
        self._error                 = None

    @property
    def device_serial_number(self) -> Optional[str]:
        return self._device_serial_number

    @property
    def name(self):
        return self._name
    
    @property
    def command_endpoint(self) -> str:
        """The endpoint which processes action commands"""
        return f"{CULLIGAN_IOT_URL:s}/device/command"
    
    def set_command_payload(self, command: str, active: Optional[bool], duration: Optional[int]=60) -> dict:
        """"""
        if command in self._commands:
            payload = {
                "command": command,
                "serialNumber": self.device_serial_number,
                "protocolVersion": 1
            }

            # only telemetry doesn't need params
            # otherwise it needs active: 0 or active 1
            if command != "telemetry.get":
                payload["params"] = {
                    "active": int(active)
                }

                if command == "bypass.timed.on":
                    payload["params"]["duration"] = duration

            return payload
    
    @property
    def all_properties_endpoint(self) -> str:
        """
            API endpoint to fetch updated device information
            This API retrieves all the properties for a specified device serial number (DSN).
        """
        return f'{CULLIGAN_IOT_URL}/device/data?serialNumber={self.device_serial_number}'
    
    def get_property_value(self, property_name: PropertyName) -> Any:
        """Get the value of a property from the properties dictionary"""
        if isinstance(property_name, Enum):
            property_name = property_name.value
        return self.properties[property_name]

    def update(self, property_list: Optional[Iterable[str]] = None):
        """Update the known device state from all properties and call _do_update to add the properties to the object property dictionary.
            Culligan returns all properties with each request."""
        full_update = True # property_list is None

        resp = self.culligan_api.self_request('get', self.all_properties_endpoint, params=None)
        properties = resp.json()
        
        return self._do_update(full_update, properties)

    async def async_update(self, property_list: Optional[Iterable[str]] = None):
        """Async update the known device state from all properties and call _do_update to add the properties to the object property dictionary
            Culligan returns all properties with each request."""
        full_update = True # property_list is None

        async with await self.culligan_api.async_request('get', self.all_properties_endpoint, params=None) as resp:
            properties = await resp.json()

        # _do_update should not be thread blocking
        return self._do_update(full_update, properties)

    def _do_update(self, full_update: bool, properties: Dict):
        """
            Update the internal state from fetched properties. This is called within update()/async_update()
            Categorize properties by Access (e.g. SET-able or Read-Only).
                Alternatively? This could use the read_only property bool instead of 'set_'.
        """

        # Update the property map so we can update by name instead of by fickle number
        if full_update:
            # Did a full update, so let's wipe everything
            self.properties = {}

        # Culligan properties are just a simple dict, no list of dicts
        datapoints = properties["data"]["datapoints"]

        for top_key in datapoints.keys():
            self.properties[top_key] = datapoints[top_key]

        # the nested datapoints don't seem interesting at the moment

        return True

    
class CulliganIoTSoftener(CulliganIoTDevice):
    """ Extend device into a water softener specific device """

    def __init__(self, culligan_api: "CulliganApi", device_dct: Dict):
        super().__init__(culligan_api, device_dct)

        self._model                     = device_dct["model"]
        self._generation                = device_dct["generation"]
        self._software_version          = device_dct["swVersion"]
        self._region                    = device_dct["region"]["code"]

        self.is_online                  = bool(device_dct["status"]["connection"]["online"])

        self._commands                  = ["telemetry.get", "awayMode.set","bypass.timed.on","bypass.permanent.on","bypass.off"]

    @property
    def device_model_number(self) -> Optional[str]:
        return self._model

    @property
    def generation(self):
        return self._generation
    
    @property
    def software_version(self):
        return self._software_version
    
    @property
    def region(self):
        return self._region
    
    def get_telemetry(self):
        """Send telemetry command"""
        resp = self.culligan_api.self_request('post', self.command_endpoint, json=self.set_command_payload("telemetry.get", True))
        json = resp.json()
        if json["success"] == True:
            return True
        else:
            return False
        
    async def async_get_telemetry(self):
        """Send telemetry command"""
        resp = await self.culligan_api.async_request('post', self.command_endpoint, json=self.set_command_payload("telemetry.get", True))
        json = await resp.json()
        if json["success"] == True:
            return True
        else:
            return False
        
    def start_vacation_mode(self):
        """Send telemetry command"""
        resp = self.culligan_api.self_request('post', self.command_endpoint, json=self.set_command_payload("awayMode.set", True))
        json = resp.json()
        if json["success"] == True:
            return True
        else:
            return False
        
    async def async_start_vacation_mode(self):
        """Send telemetry command"""
        resp = await self.culligan_api.async_request('post', self.command_endpoint, json=self.set_command_payload("awayMode.set", True))
        json = await resp.json()
        if json["success"] == True:
            return True
        else:
            return False
        
    def stop_vacation_mode(self):
        """Send telemetry command"""
        resp = self.culligan_api.self_request('post', self.command_endpoint, json=self.set_command_payload("awayMode.set", False))
        json = resp.json()
        if json["success"] == True:
            return True
        else:
            return False
        
    async def async_stop_vacation_mode(self):
        """Send telemetry command"""
        resp = await self.culligan_api.async_request('post', self.command_endpoint, json=self.set_command_payload("awayMode.set", False))
        json = await resp.json()
        if json["success"] == True:
            return True
        else:
            return False
    
    def start_bypass_mode(self):
        """Send telemetry command"""
        resp = self.culligan_api.self_request('post', self.command_endpoint, json=self.set_command_payload("bypass.permanent.on", True))
        json = resp.json()
        if json["success"] == True:
            return True
        else:
            return False
        
    async def async_start_bypass_mode(self):
        """Send telemetry command"""
        resp = await self.culligan_api.async_request('post', self.command_endpoint, json=self.set_command_payload("bypass.permanent.on", True))
        json = await resp.json()
        if json["success"] == True:
            return True
        else:
            return False
        
    def start_bypass_timed_mode(self, duration: int=60):
        """Send telemetry command"""
        resp = self.culligan_api.self_request('post', self.command_endpoint, json=self.set_command_payload("bypass.timed.on", True, duration))
        json = resp.json()
        if json["success"] == True:
            return True
        else:
            return False
        
    async def async_start_bypass_timed_mode(self, duration: int=60):
        """Send telemetry command"""
        resp = await self.culligan_api.async_request('post', self.command_endpoint, json=self.set_command_payload("bypass.permanent.on", True, duration))
        json = await resp.json()
        if json["success"] == True:
            return True
        else:
            return False
        
    def stop_bypass_mode(self):
        """Send telemetry command"""
        resp = self.culligan_api.self_request('post', self.command_endpoint, json=self.set_command_payload("bypass.off", True))
        json = resp.json()
        if json["success"] == True:
            return True
        else:
            return False
        
    async def async_stop_bypass_mode(self):
        """Send telemetry command"""
        resp = await self.culligan_api.async_request('post', self.command_endpoint, json=self.set_command_payload("bypass.off", True))
        json = await resp.json()
        if json["success"] == True:
            return True
        else:
            return False


"""Python API for Culligan devices"""

from .uniapi_culliganiot import new_culligan_api, CulliganApi
from .exc import (
    CulliganError,
    CulliganAuthExpiringError,
    CulliganNotAuthedError,
    CulliganAuthError,
    CulliganReadOnlyPropertyError,
)

__version__ = '1.1.3'



"""Exceptions"""

# Default messages
AUTH_EXPIRED_MESSAGE = 'Ayla Networks API authentication expired.  Re-authenticate and retry.'
AUTH_FAILURE_MESSAGE = 'Error authenticating to Ayla Networks.'
NOT_AUTHED_MESSAGE = 'Ayla Networks API not authenticated.  Authenticate first and retry.'


class CulliganError(RuntimeError):
    """Parent class for all Culligan exceptions"""


class CulliganAuthError(CulliganError):
    """Exception authenticating"""
    def __init__(self, msg=AUTH_FAILURE_MESSAGE, *args):
        super().__init__(msg, *args)


class CulliganAuthExpiringError(CulliganError):
    """Authentication expired and needs to be refreshed"""
    def __init__(self, msg=AUTH_EXPIRED_MESSAGE, *args):
        super().__init__(msg, *args)


class CulliganNotAuthedError(CulliganError):
    """Shark not authorized"""
    def __init__(self, msg=NOT_AUTHED_MESSAGE, *args):
        super().__init__(msg, *args)


class CulliganReadOnlyPropertyError(CulliganError):
    """Tried to set a read-only property"""
    pass


from pathlib import Path
from setuptools import setup, find_packages
from re import findall, M

packages = ["culligan"]

# The directory containing this file
HERE = Path(__file__).parent.resolve()

# The text of the README file
README = (HERE / "README.md").read_text("utf-8")

# The text of the LICENSE file
LICENSE = (HERE / "LICENSE").read_text("utf-8")

# Pull the version from __init__.py so we don't need to maintain it in multiple places
init_txt = (HERE / "src" / packages[0] / "__init__.py").read_text("utf-8")
try:
    version = findall(r"^__version__ = ['\"]([^'\"]+)['\"]\r?$", init_txt, M)[0]
except IndexError:
    raise RuntimeError('Unable to determine version.')


setup(
    name="culligan",
    version=version,
    description="Python API for Culligan products",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/rewardone/culligan",
    project_urls={
        "Bug Tracker": "https://github.com/rewardone/culligan/issues"
    },
    author="Reward One",
    author_email="rewardone@gmail.com",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    package_dir={'':"src"},
    packages=find_packages("src"),
    include_package_data=False
)